/*
 * fsck.viva.c - a file system consistency checker for Linux.
 *
 * (c) 1996. Shankar Pasupathy .
 *           Modified by Vamsi Krishna 1997.
 * 
 * A whole lot of code was borrowed from fsck.minix.c 
 * (C) 1991, 1992 Linus Torvalds. 
 *
 * This file may be redistributed as per the GNU copyleft.
 *	
 */

/*
 * Usuage: fsck [-larvsm] device
 *	-l for a listing of all the filenames
 *	-a for automatic repairs (not implemented)
 *	-r for repairs (interactive) (not implemented)
 *	-v for verbose (tells how many files)
 *	-s for super-block info
 *	-m for viva-like "mode not cleared" warnings
 *	-f force filesystem check even if filesystem marked as valid
 *
 * The device may be a block device or a image of one, but this isn't
 * enforced (but it's not much fun on a character device :-). 
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <ctype.h>
#include <stdlib.h>
#include <termios.h>
#include <mntent.h>
#include <sys/stat.h>

#include "fs.h"
#include "../../viva_fs.h"

#ifndef __GNUC__
#error "needs gcc for the bitop-__asm__'s"
#endif

#ifndef __linux__
#define volatile
#endif

#include <asm/system.h>
#include <asm/segment.h>


/*
 * Some hacks to defeat gcc over-optimizations..
 */
struct __dummy { unsigned long a[100]; };
#define ADDR (*(struct __dummy *) addr)

/*
 * Find-bit routines..
 */
static inline int find_first_zero_bit(void * addr, unsigned size)
{
	int res;

	if (!size)
		return 0;
	__asm__("
		cld
		movl $-1,%%eax
		repe; scasl
		je 1f
		subl $4,%%edi
		movl (%%edi),%%eax
		notl %%eax
		bsfl %%eax,%%edx
		jmp 2f
1:		xorl %%edx,%%edx
2:		subl %%ebx,%%edi
		shll $3,%%edi
		addl %%edi,%%edx"
		:"=d" (res)
		:"c" ((size + 31) >> 5), "D" (addr), "b" (addr)
		:"ax", "bx", "cx", "di");
	return res;
}

static inline int find_next_zero_bit (void * addr, int size, int offset)
{
	unsigned long * p = ((unsigned long *) addr) + (offset >> 5);
	int set = 0, bit = offset & 31, res;
	
	if (bit) {
		/*
		 * Look for zero in first byte
		 */
		__asm__("
			bsfl %1,%0
			jne 1f
			movl $32, %0
1:			"
			: "=r" (set)
			: "r" (~(*p >> bit)));
		if (set < (32 - bit))
			return set + offset;
		set = 32 - bit;
		p++;
	}
	/*
	 * No zero yet, search remaining full bytes for a zero
	 */
	res = find_first_zero_bit (p, size - 32 * (p - (unsigned long *) addr));
	return (offset + set + res);
}

#define count_bits_set(byte) (bits_set[*(unsigned char *)(byte)])

unsigned char bits_set[256] = { /* number of 1s in a byte */
    0,1,1,2,1,2,2,3,1,2,2,3,2,3,3,4,
    1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
    1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
    2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
    1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
    2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
    2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
    3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
    1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
    2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
    2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
    3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
    2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
    3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
    3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
    4,5,5,6,5,6,6,7,5,6,6,7,6,7,7,8
};

/* RETURN 0 on error */
unsigned long get_zone_addr(unsigned char *map,int nr)
{
int i,block_nr,bit_nr,byte; /* nr stands for number */
unsigned char temp_byte[1];
struct viva_smap_entry *entry; /* defined in viva_fs.h */

if (nr<0) {
	return 0;
}
entry = (struct viva_smap_entry *) (map);
block_nr = -1;
while (entry->start_addr) {
	block_nr++; /* because the first bit is implicitly set */
	/* and for the same reason check the following: */
	if (nr == block_nr) {
		return(entry->start_addr);
	}
	byte = 0;
	while (byte < entry->length) {
		if (nr < (block_nr + count_bits_set(entry->bitmap + byte)+1)) 
			goto block_in_this_byte;
		block_nr += count_bits_set(entry->bitmap + byte);
		byte++;
	}
/* it might be in the next entry */
entry = (struct viva_smap_entry *) (entry->bitmap + entry->length);
}
/* block requested in not available - error !*/
return 0;

block_in_this_byte:
	memcpy(temp_byte,(entry->bitmap+byte),1);
	*temp_byte = *temp_byte ^ 0xff; /* make all 0s, 1s */
	bit_nr = -1;
	for (i=0; i<(nr-block_nr); i++) {
		bit_nr = find_next_zero_bit(temp_byte,8,bit_nr+1);
	}
	if (bit_nr > 7) {
		return 0; /* error */
	}
	return(entry->start_addr + byte * 8 + bit_nr + 1);
}

#define ROOT_INO 1

#define UPPER(size,n) ((size+((n)-1))/(n))
#define INODE_SIZE (sizeof(struct viva_inode))
#define INODE_BLOCKS UPPER(INODES,VIVA_INODES_PER_BLOCK)
#define INODE_BUFFER_SIZE (INODE_BLOCKS * VIVA_BLOCK_SIZE)

#define BITS_PER_BLOCK (VIVA_BLOCK_SIZE<<3)

static char * program_name = "fsck.viva";
static char * program_version = "1.0 - 07/08/96";
static char * device_name = NULL;
static int IN;
static int repair=0, automatic=0, verbose=0, list=0, show=0, warn_mode=0, 
	force=0;
static int directory=0, regular=0, blockdev=0, chardev=0, links=0,
		symlinks=0, total=0;

static int changed = 0; /* flags if the filesystem has been changed */
static int errors_uncorrected = 0; /* flag if some error was not corrected */
static int dirsize = 16;
static int namelen = 14;

/* File-name data */
#define MAX_DEPTH 50
static int name_depth = 0;
static char name_list[MAX_DEPTH][NAME_MAX+1];

static char * inode_buffer = NULL;
#define Inode (((struct viva_inode *) inode_buffer)-1)
static char super_block_buffer[VIVA_BLOCK_SIZE];
#define Super (*(struct viva_super_block *)super_block_buffer)
#define INODES ((unsigned long)Super.s_ninodes)
#define ZONES ((unsigned long)Super.s_nzones)
#define IMAPS ((unsigned long)Super.s_imap_blocks)
#define ZMAPS ((unsigned long)Super.s_zmap_blocks)
#define FS_MAPS ((unsigned long)Super.s_fsmap_blocks)
#define FA_MAPS ((unsigned long)Super.s_famap_blocks)
#define FIRSTZONE ((unsigned long)Super.s_firstdatazone)
#define ZONESIZE ((unsigned long)Super.s_log_zone_size)
#define MAGIC (Super.s_magic)
#define NORM_FIRSTZONE (2+IMAPS+ZMAPS+FS_MAPS+FA_MAPS+INODE_BLOCKS)

static char inode_map[VIVA_BLOCK_SIZE * VIVA_I_MAP_SLOTS];
static char zone_map[VIVA_BLOCK_SIZE * VIVA_Z_MAP_SLOTS];
static char frag_map[VIVA_FS_MAP_SIZE];
static char frag_alloc_map[VIVA_BLOCK_SIZE * VIVA_FA_MAP_SLOTS];


static unsigned char * inode_count = NULL;
static unsigned char * zone_count = NULL;
static unsigned char * frag_count = NULL;

void recursive_check(unsigned int ino);

#define bitop(name,op) \
static inline int name(char * addr,unsigned int nr) \
{ \
int __res; \
__asm__ __volatile__("bt" op "l %1,%2; adcl $0,%0" \
:"=g" (__res) \
:"r" (nr),"m" (*(addr)),"0" (0)); \
return __res; \
}

bitop(bit,"")
bitop(setbit,"s")
bitop(clrbit,"r")

#define inode_in_use(x) (bit(inode_map,(x)))
#define zone_in_use(x) (bit(zone_map,(x)-FIRSTZONE))
#define frag_block_in_use(x) (bit(frag_map,(x)-FIRSTZONE))

#define mark_inode(x) (setbit(inode_map,(x)),changed=1)
#define unmark_inode(x) (clrbit(inode_map,(x)),changed=1)

#define mark_zone(x) (setbit(zone_map,(x)-FIRSTZONE),changed=1)
#define unmark_zone(x) (clrbit(zone_map,(x)-FIRSTZONE),changed=1)

#define mark_frag_block(x) (setbit(frag_map,(x)-FIRSTZONE))
#define unmark_frag_block(x) (clrbit(frag_map,(x)-FIRSTZONE))

#define mark_frag_alloc(x) (setbit(frag_alloc_map, ((x)- (FIRSTZONE*VIVA_FRAG_PER_BLOCK))))
#define unmark_frag_alloc(x) (clrbit(frag_alloc_map, ((x)- (FIRSTZONE*VIVA_FRAG_PER_BLOCK))))


/*
 * Volatile to let gcc know that this doesn't return. When trying
 * to compile this under viva, volatile gives a warning, as
 * exit() isn't defined as volatile under viva.
 */
volatile void fatal_error(const char * fmt_string, int status)
{
	fprintf(stderr,fmt_string,program_name,device_name);
	exit(status);
}

#define usage() fatal_error("Usage: %s [-larvsmf] /dev/name\n",16)
#define die(str) fatal_error("%s: " str "\n",8)

/*
 * This simply goes through the file-name data and prints out the
 * current file.
 */
void print_current_name(void)
{
	int i=0;

	while (i<name_depth)
		printf("/%.*s",namelen,name_list[i++]);
}

int ask(const char * string,int def)
{
	int c;

	if (!repair) {
		printf("\n");
		errors_uncorrected = 1;
		return 0;
	}
	if (automatic) {
		printf("\n");
		if (!def)
		      errors_uncorrected = 1;
		return def;
	}
	printf(def?"%s (y/n)? ":"%s (n/y)? ",string);
	for (;;) {
		fflush(stdout);
		if ((c=getchar())==EOF) {
		        if (!def)
			      errors_uncorrected = 1;
			return def;
		}
		c=toupper(c);
		if (c == 'Y') {
			def = 1;
			break;
		} else if (c == 'N') {
			def = 0;
			break;
		} else if (c == ' ' || c == '\n')
			break;
	}
	if (def)
		printf("y\n");
	else {
		printf("n\n");
		errors_uncorrected = 1;
	     }
	return def;
}

/*
 * Make certain that we aren't checking a filesystem that is on a
 * mounted partition.  Code adapted from e2fsck, Copyright (C) 1993,
 * 1994 Theodore Ts'o.  Also licensed under GPL.
 */
static void check_mount(void)
{
	FILE * f;
	struct mntent * mnt;
	int cont;
	int fd;

	if ((f = setmntent (MOUNTED, "r")) == NULL)
		return;
	while ((mnt = getmntent (f)) != NULL)
		if (strcmp (device_name, mnt->mnt_fsname) == 0)
			break;
	endmntent (f);
	if (!mnt)
		return;

	/*
	 * If the root is mounted read-only, then /etc/mtab is
	 * probably not correct; so we won't issue a warning based on
	 * it.
	 */
	fd = open(MOUNTED, O_RDWR);
	if (fd < 0 && errno == EROFS)
		return;
	else
		close(fd);
	
	printf ("%s is mounted.	 ", device_name);
	if (isatty(0) && isatty(1))
/*		cont = ask("Do you really want to continue", 0);*/
		cont = 1;
	else
		cont = 1;
	if (!cont) {
		printf ("check aborted.\n");
		exit (0);
	}
	return;
}

/*
 * check_zone_nr checks to see that *nr is a valid zone nr. 
 */
unsigned long check_zone_nr(unsigned long nr)
{
	if (!nr)
		return 0;
	if (nr < FIRSTZONE)
		printf("Zone nr < FIRSTZONE in file `");
	else if (nr >= ZONES)
		printf("Zone nr %d >= ZONES in file `",nr);
	else
		return nr;
	return 0;
}

/*
 * read-block reads block nr into the buffer at addr.
 */
void read_block(unsigned int nr, char * addr)
{
	if (!nr) {
		memset(addr,0,VIVA_BLOCK_SIZE);
		return;
	}
	if (VIVA_BLOCK_SIZE*nr != lseek(IN, VIVA_BLOCK_SIZE*nr, SEEK_SET)) {
		printf("Read error: unable to seek to block in file '");
		print_current_name();
		printf("'\n");
		memset(addr,0,VIVA_BLOCK_SIZE);
		errors_uncorrected = 1;
	} else {
		if (VIVA_BLOCK_SIZE != read(IN, addr, VIVA_BLOCK_SIZE)) {
		printf("Read error: bad block in file '");
		print_current_name();
		printf("'\n");
		memset(addr,0,VIVA_BLOCK_SIZE);
		errors_uncorrected = 1;
	}
	}
}

/*
 * write_block writes block nr to disk.
 */
void write_block(unsigned int nr, char * addr)
{
	if (!nr)
		return;
	if (nr < FIRSTZONE || nr >= ZONES) {
		printf("Internal error: trying to write bad block\n"
		"Write request ignored\n");
		errors_uncorrected = 1;
		return;
	}
	if (VIVA_BLOCK_SIZE*nr != lseek(IN, VIVA_BLOCK_SIZE*nr, SEEK_SET))
		die("seek failed in write_block");
	if (VIVA_BLOCK_SIZE != write(IN, addr, VIVA_BLOCK_SIZE)) {
		printf("Write error: bad block in file '");
		print_current_name();
		printf("'\n");
		errors_uncorrected = 1;
	}
}

/*
 * map-block calculates the absolute block nr of a block in a file.
 * It sets 'changed' if the inode has needed changing.
 */
/* CHANGE HERE for viva */
int map_block(struct viva_inode * inode, unsigned int blknr)
{
	int blk_chg, block, result;
	unsigned char sparse_bitmap[VIVA_SMAP_SIZE];
	static char buffer[VIVA_BLOCK_SIZE];
	unsigned long block_addr;
		
	memcpy(sparse_bitmap,inode->i_smap,VIVA_DIRECT_SMAP_SIZE);
	if (inode->i_indirect_block) {
		read_block((int)inode->i_indirect_block,buffer);
		memcpy(sparse_bitmap+VIVA_DIRECT_SMAP_SIZE,buffer,VIVA_BLOCK_SIZE);
	}
	block_addr = get_zone_addr(sparse_bitmap,blknr);
	result = check_zone_nr(block_addr);
	return result;
}

void write_super_block(void)
{
	/*
	 * Set the state of the filesystem based on whether or not there
	 * are uncorrected errors.  The filesystem valid flag is
	 * unconditionally set if we get this far.
	 */
	Super.s_state |= VIVA_VALID_FS;
	if ( errors_uncorrected )
		Super.s_state |= VIVA_ERROR_FS;
	else
		Super.s_state &= ~VIVA_ERROR_FS;
	
	if (VIVA_BLOCK_SIZE != lseek(IN, VIVA_BLOCK_SIZE, SEEK_SET))
		die("seek failed in write_super_block");
	if (VIVA_BLOCK_SIZE != write(IN, super_block_buffer, VIVA_BLOCK_SIZE))
		die("unable to write super-block");

	return;
}

void write_tables(void)
{
	write_super_block();

	if (IMAPS*VIVA_BLOCK_SIZE != write(IN,inode_map,IMAPS*VIVA_BLOCK_SIZE))
		die("Unable to write inode map");
	if (ZMAPS*VIVA_BLOCK_SIZE != write(IN,zone_map,ZMAPS*VIVA_BLOCK_SIZE))
		die("Unable to write zone map");
	if (FS_MAPS*VIVA_BLOCK_SIZE != write(IN,frag_map,FS_MAPS*VIVA_BLOCK_SIZE))
		die("Unable to write frag_sparse map");
	if (FA_MAPS*VIVA_BLOCK_SIZE != write(IN,frag_alloc_map,FA_MAPS*VIVA_BLOCK_SIZE))
		die("Unable to write frag_alloc map");
	if (INODE_BUFFER_SIZE != write(IN,inode_buffer,INODE_BUFFER_SIZE))
		die("Unable to write inodes");
}

void read_tables(void)
{
	int temp;
	memset(inode_map,0,sizeof(inode_map));
	memset(zone_map,0,sizeof(zone_map));
	if (VIVA_BLOCK_SIZE != lseek(IN, VIVA_BLOCK_SIZE, SEEK_SET))
		die("seek failed");
	if (VIVA_BLOCK_SIZE != (temp=read(IN, super_block_buffer, VIVA_BLOCK_SIZE)))
		die("unable to read super block ");
	if (MAGIC == VIVA_SUPER_MAGIC) {
		namelen = 14;
		dirsize = 16;
	} else if (MAGIC == VIVA_SUPER_MAGIC2) {
		namelen = 30;
		dirsize = 32;
	} else
		die("bad magic number in super-block");
	/* very important - check to see if block size = 1024, since viva
	won't work with larger block sizes right now. However, that can
	be fixed very easily */
	/* Vamsi: Check removed temporarily  */
/*	if (ZONESIZE != 0 || VIVA_BLOCK_SIZE != 1024)  */
/*		die("Only 1k blocks/zones supported");     */
	if (!IMAPS || IMAPS > VIVA_I_MAP_SLOTS)
		die("No inode bitmaps or excessive number of them");
	if (!ZMAPS || ZMAPS > VIVA_Z_MAP_SLOTS)
		die("No block bitmaps or excessive number of them");
	if (!FS_MAPS || FS_MAPS > VIVA_FS_MAP_SLOTS)
		die("No frag_sparse bitmaps or excessive number of them");
	if (!FA_MAPS || FA_MAPS > VIVA_FA_MAP_SLOTS)
		die("No frag_alloc bitmaps or excessive number of them");
	inode_buffer = malloc(INODE_BUFFER_SIZE);
	if (!inode_buffer)
		die("Unable to allocate buffer for inodes");
	inode_count = malloc(INODES);
	if (!inode_count)
		die("Unable to allocate buffer for inode count");
	zone_count = malloc(ZONES);
	if (!zone_count)
		die("Unable to allocate buffer for block count");
	frag_count = malloc(ZONES*VIVA_FRAG_PER_BLOCK);
	if (!frag_count)
		die("Unable to allocate buffer for frag count");
	if (IMAPS*VIVA_BLOCK_SIZE != read(IN,inode_map,IMAPS*VIVA_BLOCK_SIZE))
		die("Unable to read inode bitmap");
	if (ZMAPS*VIVA_BLOCK_SIZE != read(IN,zone_map,ZMAPS*VIVA_BLOCK_SIZE))
		die("Unable to read block bitmap");
	if (FS_MAPS*VIVA_BLOCK_SIZE != read(IN,frag_map,FS_MAPS*VIVA_BLOCK_SIZE))
		die("Unable to read frag_sparse bitmap");
	if (FA_MAPS*VIVA_BLOCK_SIZE != read(IN,frag_alloc_map,FA_MAPS*VIVA_BLOCK_SIZE))
		die("Unable to read frag_alloc bitmap");
	if (INODE_BUFFER_SIZE != read(IN,inode_buffer,INODE_BUFFER_SIZE))
		die("Unable to read inodes");
	if (NORM_FIRSTZONE != FIRSTZONE) {
		printf("Warning: First data block read is not in the expected place\n");
		errors_uncorrected = 1;
	}
	if (show) {
		printf("%d inodes\n",INODES);
		printf("%d blocks\n",ZONES);
		printf("First data block=%d (Normal = %d)\n",FIRSTZONE,NORM_FIRSTZONE);
		printf("Block size in bytes = %d\n",VIVA_BLOCK_SIZE<<ZONESIZE);
		printf("Filesystem state=%d\n", Super.s_state);
	}
}

struct viva_inode * get_inode(unsigned int nr)
{
	struct viva_inode * inode;

	if (!nr || nr >= INODES)
		return NULL;
	total++;
	inode = Inode + nr;
	if (!inode_count[nr]) {
		if (!inode_in_use(nr)) {
			printf("Inode %d marked not used, but used for file '",
				nr);
			print_current_name();
			printf("'\n");
			if (repair)
				if (ask("Mark in use",1))
					mark_inode(nr);
			else
			        errors_uncorrected = 1;
		}
		if (S_ISDIR(inode->i_mode))
			directory++;
		else if (S_ISREG(inode->i_mode))
			regular++;
		else if (S_ISCHR(inode->i_mode))
			chardev++;
		else if (S_ISBLK(inode->i_mode))
			blockdev++;
		else if (S_ISLNK(inode->i_mode))
			symlinks++;
		else if (S_ISSOCK(inode->i_mode))
			;
		else if (S_ISFIFO(inode->i_mode))
			;
		else {
                        print_current_name();
                        printf(" has mode %05o\n",inode->i_mode);
                }

	} else
		links++;
	if (!++inode_count[nr]) {
		printf("Warning: inode count too big.\n");
		inode_count[nr]--;
		errors_uncorrected = 1;
	}
	return inode;
}

void check_root(void)
{
	struct viva_inode * inode = Inode + ROOT_INO;

	if (!inode || !S_ISDIR(inode->i_mode))
		die("root inode isn't a directory");
}

static int add_zone(unsigned long znr, int * corrected)
{
	/* znr stands for zone (block) number */
	int result;
	unsigned long block;

	result = 0;
	/* Check if znr is a valid zone, i.e. if it lies between the
	first data zone and last data zone */
	block = check_zone_nr(znr);
	if (!block)
	{
		printf("Vamsi: Due to add_zone \n");
		return 0;
	}
	if (zone_count[block]) {
		printf("Block %d has been used before. Now in file `",block);
		print_current_name();
		printf("'.");
		printf("Can't fix this right now. Delete the file\n");
	}
	if (!block)
		return 0;
	if (!zone_in_use(block)) {
		printf("Block %d in file `",block);
		print_current_name();
		printf("' is marked not in use.");
		if (ask("Correct",1))
			mark_zone(block);
	}
	if (!++zone_count[block])
		zone_count[block]--;
	return block;
}

static int add_frag(unsigned long frag_block_nr, short frag_nr)
{
	/* frag_block_nr stands for Frag_block number */
	int result;
	unsigned long block;

	result = 0;
	/* Check if frag_block_nr is a valid zone, i.e. if it lies between the
	first data zone and last data zone */
	block = check_zone_nr(frag_block_nr);
	if (!block)
	{
		printf("Vamsi: Due to add_frag \n");
		return 0;
	}
	if (frag_count[(block*VIVA_FRAG_PER_BLOCK)+frag_nr]) {
		printf("Fragment %d of block %d has been used before. Now in file `",
																frag_nr,block);
		print_current_name();
		printf("'.");
		printf("Can't fix this right now. Delete the file\n");
	}
	if (!block)
		return 0;
/*
	if (!frag_block_in_use(block)) {
		printf("Frag_block %d in file `",block);
		print_current_name();
		printf("' is marked not used in frag_map.");
		if (ask("Correct",1))
			mark_frag_block(block);
	}
*/
	if (!zone_in_use(block))
	{
		printf("Frag_block %d in file `",block);
		print_current_name();
		printf("' is marked not used in zone_map.");
		if (ask("Correct",1))
			mark_zone(block);
	}
	if (!frag_count[(block*VIVA_FRAG_PER_BLOCK)+frag_nr])
		frag_count[(block*VIVA_FRAG_PER_BLOCK)+frag_nr]++;
	return block;
}

void check_zones(unsigned int i)
{
	struct viva_inode * inode;
	unsigned char sparse_bitmap[VIVA_SMAP_SIZE];
	unsigned char smap_buffer[VIVA_BLOCK_SIZE];
	unsigned long zone_addr;	

	if (!i || i >= INODES)
		return;
	if (inode_count[i] > 1)	/* have we counted this file already? */
		return;
	inode = Inode + i;
	if (!S_ISDIR(inode->i_mode) && !S_ISREG(inode->i_mode) &&
	    !S_ISLNK(inode->i_mode))
		return;
	
	memcpy(sparse_bitmap,inode->i_smap,VIVA_DIRECT_SMAP_SIZE);
	if (inode->i_indirect_block) {
		/* read in the indirect block */
		read_block(inode->i_indirect_block,smap_buffer);
		memcpy(sparse_bitmap+VIVA_DIRECT_SMAP_SIZE,smap_buffer,VIVA_BLOCK_SIZE);
		add_zone(inode->i_indirect_block,&changed);
	}
	i = 0;
	while(zone_addr = get_zone_addr(sparse_bitmap,i)) {
		add_zone(zone_addr, &changed);
		i++;
	}
}

void check_file(struct viva_inode * dir, unsigned int offset)
{
	static char blk[VIVA_BLOCK_SIZE];
	struct viva_inode * inode;
	int ino;
	char * name;
	unsigned int block;
	int count;

	block = map_block(dir,offset/VIVA_BLOCK_SIZE);
	read_block(block, blk);
	name = blk + (offset % VIVA_BLOCK_SIZE) + 2;
	ino = * (unsigned short *) (name-2);
	if (ino >= INODES) {
		print_current_name();
		printf(" contains a bad inode number for file '");
		printf("%.*s'.",namelen,name);
		if (ask(" Remove",1)) {
			*(unsigned short *)(name-2) = 0;
			write_block(block, blk);
		}
		ino = 0;
	}	
	inode = get_inode(ino);
	if (!offset)
		if (!inode || strcmp(".",name)) {
			print_current_name();
			printf(": bad directory: '.' isn't first\n");
			errors_uncorrected = 1;
		} else return;
	if (offset == dirsize)
		if (!inode || strcmp("..",name)) {
			print_current_name();
			printf(": bad directory: '..' isn't second\n");
			errors_uncorrected = 1;
		} else return;
	if (!inode)
		return;
	if (name_depth < MAX_DEPTH)
		strncpy(name_list[name_depth],name,namelen);
	name_depth++;	
	if (list) {
		if (verbose)
			printf("%6d %07o %3d ",ino,inode->i_mode,inode->i_nlinks);
		print_current_name();
		if (S_ISDIR(inode->i_mode))
			printf(":\n");
		else
			printf("\n");
	}
	check_zones(ino);
	if(inode && inode->i_fragsInUse)
		for(count=0; count < inode->i_fragsInUse; count++)
			add_frag(inode->i_frag_block, inode->i_frag_no[count]);
	if (inode && S_ISDIR(inode->i_mode))
		recursive_check(ino);
	name_depth--;
	return;
}

void recursive_check(unsigned int ino)
{
	struct viva_inode * dir;
	unsigned int offset;

	dir = Inode + ino;
	if (!S_ISDIR(dir->i_mode))
		die("internal error");
	if (dir->i_size < 32) {
		print_current_name();
		printf(": bad directory: size<32");
		errors_uncorrected = 1;
	}
	for (offset = 0 ; offset < dir->i_size ; offset += dirsize)
		check_file(dir,offset);
}

int bad_zone(int i)
{
	char buffer[VIVA_BLOCK_SIZE];

	if (VIVA_BLOCK_SIZE*i != lseek(IN, VIVA_BLOCK_SIZE*i, SEEK_SET))
		die("seek failed in bad_zone");
	return (VIVA_BLOCK_SIZE != read(IN, buffer, VIVA_BLOCK_SIZE));
}

void check_counts(void)
{
	int i;
	int count, temp;

	for (i=1 ; i < INODES ; i++) {
		/*
		if (!inode_in_use(i) && Inode[i].i_mode && warn_mode) {
			printf("Inode %d mode not cleared. %07o",i,Inode[i].i_mode);
			if (ask("Clear",1)) {
				Inode[i].i_mode = 0;
				changed = 1;
			}
		}
		*/
		if (!inode_count[i]) {
			if (!inode_in_use(i))
				continue;
			printf("Inode %d not used, marked used in the bitmap.",i);
			if (ask("Clear",1))
				unmark_inode(i);
			continue;
		}
		if (!inode_in_use(i)) {
			printf("Inode %d used, marked unused in the bitmap.",
				i);
			if (ask("Set",1))
				mark_inode(i);
		}
		if (Inode[i].i_nlinks != inode_count[i]) {
			printf("Inode %d (mode = %07o), i_nlinks=%d, counted=%d.",
				i,Inode[i].i_mode,Inode[i].i_nlinks,inode_count[i]);
			if (ask("Set i_nlinks to count",1)) {
				Inode[i].i_nlinks=inode_count[i];
				changed=1;
			}
		}
	}
	for (i=FIRSTZONE ; i < ZONES ; i++) {
		if ( zone_in_use(i)==zone_count[i] ) 
			continue;
		if( zone_in_use(i)==frag_block_in_use(i) )
			continue;
		if (!zone_count[i]) {
			if (bad_zone(i))
				continue;
/*
			temp = 0;
			for(count=0; count<VIVA_FRAG_PER_BLOCK; count++)
				temp += frag_count[(i*VIVA_FRAG_PER_BLOCK)+count];
			if(!temp)
				printf("Block %d: marked in use %d (%d), no file uses it.",i,zone_in_use(i),frag_block_in_use(i));
*/
			if(repair && (ask("Unmark",1)))
				unmark_zone(i);
			continue;
		}
		printf("Block %d: %sin use, counted=%d\n",
		i,zone_in_use(i)?"":"not ",zone_count[i]);
	}
}

void check(void)
{
	memset(inode_count,0,INODES*sizeof(*inode_count));
	memset(zone_count,0,ZONES*sizeof(*zone_count));
	memset(frag_count,0,ZONES*VIVA_FRAG_PER_BLOCK*sizeof(*frag_count));
	check_zones(ROOT_INO);
	recursive_check(ROOT_INO);
	check_counts();
}

int main(int argc, char ** argv)
{
	struct termios termios,tmp;
	int count;
	int retcode = 0;

	if (argc && *argv)
		program_name = *argv;
	if (INODE_SIZE * VIVA_INODES_PER_BLOCK != VIVA_BLOCK_SIZE)
		die("block size must be a multiple of inode size");
	while (argc-- > 1) {
		argv++;
		if (argv[0][0] != '-')
			if (device_name)
				usage();
			else
				device_name = argv[0];
		else while (*++argv[0])
			switch (argv[0][0]) {
				case 'l': list=1; break;
				case 'a': automatic=1; repair=1; break;
				case 'r': automatic=0; repair=1; break;
				case 'v': verbose=1; break;
				case 's': show=1; break;
				case 'm': warn_mode=1; break;
				case 'f': force=1; break;
				default: usage();
			}
	}
	if (!device_name)
		usage();
	check_mount();		/* trying to check a mounted filesystem? */
	/* check if repairs will be automatic or interactive */
	if (repair && !automatic) {
		if (!isatty(0) || !isatty(1))
			die("need terminal for interactive repairs");
		tcgetattr(0,&termios);
		tmp = termios;
		tmp.c_lflag &= ~(ICANON|ECHO);
		tcsetattr(0,TCSANOW,&tmp);
	}
	IN = open(device_name,repair?O_RDWR:O_RDONLY);
	if (IN < 0)
		die("unable to open '%s'");
	for (count=0 ; count<3 ; count++)
		sync(); /* why do a sync 3 times ? */
	read_tables();

	/*
	 * Determine whether or not we should continue with the checking.
	 * This is based on the status of the filesystem valid and error
	 * flags and whether or not the -f switch was specified on the 
	 * command line.
	 */
	printf("%s, %s\n", program_name, program_version);
	if ( !(Super.s_state & VIVA_ERROR_FS) && 
	      (Super.s_state & VIVA_VALID_FS) && 
	      !force ) {
		if (repair)
			printf("%s is clean, no check.\n", device_name);
		if (repair && !automatic)
			tcsetattr(0,TCSANOW,&termios);
		printf("Device %s is clean, no check necessary\n",device_name);
		return retcode;
	}
	else if (force)
		printf("Forcing filesystem check on %s.\n", device_name);
	else if (repair)
		printf("Filesystem on %s is dirty, needs checking.\n",\
			device_name);
	/* check that the root inode is in fact a directory */
if (verbose) printf("Checking root inode\n");
	check_root();
if (verbose) printf("Check completed\nChecking files recursively\n");
	check();
if (verbose) printf("Recursive check completed\n");
	if (verbose) {
		unsigned long i, free;
printf("\nSummary of the viva filesystem on device %s\n",device_name);
		for (i=1,free=0 ; i < INODES ; i++)
			if (!inode_in_use(i))
				free++;
		printf("\n%6d inodes used (%d%%)\n",(INODES-free-1),
			100*(INODES-free-1)/(INODES-1));
		for (i=FIRSTZONE,free=0 ; i < ZONES ; i++)
			if (!zone_in_use(i))
				free++;
		printf("%6d blocks used (%d%%)\n",(ZONES-free),
			100*(ZONES-free)/ZONES);
		for (i=FIRSTZONE,free=0 ; i < ZONES ; i++)
			if (!frag_block_in_use(i))
				free++;
		printf("%6d Frag_blocks used (%d%%)\n",(ZONES-free-FIRSTZONE),
			100*(ZONES-free-FIRSTZONE)/ZONES);
		printf("\n%6d regular files\n"
		"%6d directories\n"
		"%6d character device files\n"
		"%6d block device files\n"
		"%6d links\n"
		"%6d symbolic links\n"
		"    ------------\n"
		"%6d files (including directories, and links)\n",
		regular,directory,chardev,blockdev,
		links-2*directory+1,symlinks,total-2*directory+1);
	}
	if (changed) {
		write_tables();
		printf(	"----------------------------\n"
			"FILE SYSTEM HAS BEEN CHANGED\n"
			"----------------------------\n");
		for (count=0 ; count<3 ; count++)
			sync();
	}
	else if ( repair )
		write_super_block();
	
	if (repair && !automatic)
		tcsetattr(0,TCSANOW,&termios);

	if (changed)
	      retcode += 3;
	if (errors_uncorrected)
	      retcode += 4;
	return retcode;
}
