/* -----------------------------------------------------------------
The VIVA file system - version 2.0

Implemented by Shankar Pasupathy, July 1996. (shankar@pop.uky.edu)
Modified by Vamsi Krishna, July 1997. (vamsi@cslab.uky.edu)
VIVA was developed by Eric Herrin II and Raphael Finkel.
Department of Computer Science,
University of Kentucky, Lexington 40506.
Distribution site: http://www.cslab.uky.edu/~vamsi/viva

This implementation uses code from the Minix filesystem,
(Copyright: Linus Torvalds).

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation. Read the file COPYRIGHT, included with
this distribution, for further details.
--------------------------------------------------------------------*/

#ifndef _VIVA_FS_I
#define _VIVA_FS_I

/*
 * VIVA fs inode data in memory
 */
struct viva_inode_info {

	__u32   i_lastblock; 
	__u16   i_blocksInUse;
	__u8	i_cached_smap[223+2048];  

	/* i_cached_smap consists of an unsigned long address, unsigned char length
	in bytes, and sparse bitmap of the given length. The i_cached_smap may
	also contain the sparse bitmap stored in an indirect block. */

	/* 223 bytes of direct inode sparse bitmap, and 2048 bytes of indirect
	sparse bitmap. 223 is the same as VIVA_DIRECT_SMAP_SIZE and 2048 is
	the same as VIVA_INDIRECT_SMAP_SIZE */

	__u32	i_frag_block;
	__u8	i_frag_no[3];
	__u8	i_fragsInUse;

	__u32	i_cached_indirect_block;

	/* indblock_addr is a pointer to a block containing another sparse
	bitmap. This sparse bitmap has length 2048 bytes because currently
	VIVA allows only 1K block sizes. */

	struct buffer_head *indirect_smap_bh;
	
	/* buffer that is used to read in the indirect block. Also used to
	easily write out the indirect block (for e.g. when sync is called) */

	__u8   i_iblock_dirty;

	/* If i_iblock_dirty is 1, the indirect block was modified so
	it will be written out to disk */

};

#endif
