/* -----------------------------------------------------------------
The VIVA file system - version 2.0

Implemented by Shankar Pasupathy, July 1996. (shankar@pop.uky.edu)
Modified by Vamsi Krishna, July 1997. (vamsi@cslab.uky.edu)
VIVA was developed by Eric Herrin II and Raphael Finkel.
Department of Computer Science,
University of Kentucky, Lexington 40506.
Distribution site: http://www.cslab.uky.edu/~vamsi/viva

This implementation uses code from the Minix filesystem,
(Copyright: Linus Torvalds).

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation. Read the file COPYRIGHT, included with
this distribution, for further details.
--------------------------------------------------------------------*/

 /* Viva's directory handling functions - verbatim from minix/dir.c */

#include <linux/string.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/viva_fs.h>
#include <linux/stat.h>

#include <asm/segment.h>

static int viva_dir_read(struct inode * inode, struct file * filp, char * buf, int count)
{
	return -EISDIR;
}

static int viva_readdir(struct inode *, struct file *, void *, filldir_t);

static struct file_operations viva_dir_operations = {
	viva_lseek,		/* lseek - VIVA version */
	viva_dir_read,		/* read */
	NULL,			/* write - bad */
	viva_readdir,		/* readdir */
	NULL,			/* select - default */
	NULL,			/* ioctl - default */
	NULL,			/* mmap */
	NULL,			/* no special open code */
	NULL,			/* no special release code */
	file_fsync		/* default fsync */
};

/*
 * directories can handle most operations...
 */
struct inode_operations viva_dir_inode_operations = {
	&viva_dir_operations,	/* default directory file-ops */
	viva_create,		/* create */
	viva_lookup,		/* lookup */
	viva_link,		/* link */
	viva_unlink,		/* unlink */
	viva_symlink,		/* symlink */
	viva_mkdir,		/* mkdir */
	viva_rmdir,		/* rmdir */
	viva_mknod,		/* mknod */
	viva_rename,		/* rename */
	NULL,			/* readlink */
	NULL,			/* follow_link */
	NULL,			/* readpage */
	NULL,			/* writepage */
	NULL,			/* bmap */
	viva_truncate,		/* truncate */
	NULL			/* permission */
};

static int viva_readdir(struct inode * inode, struct file * filp,
	void * dirent, filldir_t filldir)
{
	unsigned int offset;
	struct buffer_head * bh;
	struct viva_dir_entry * de;
	struct viva_sb_info * info;

	if (!inode || !inode->i_sb || !S_ISDIR(inode->i_mode))
		return -EBADF;
	info = &inode->i_sb->u.viva_sb;
	if (filp->f_pos & (info->s_dirsize - 1))
		return -EBADF;
	while (filp->f_pos < inode->i_size) {
		offset = filp->f_pos & (VIVA_BLOCK_SIZE-1);
		bh = viva_bread(inode,(filp->f_pos)>>VIVA_BLOCK_SIZE_BITS,0);
		if (!bh) {
			filp->f_pos += VIVA_BLOCK_SIZE - offset;
			continue;
		}
		do {
			de = (struct viva_dir_entry *) (offset + bh->b_data);
			if (de->inode) {
				int size = strnlen(de->name, info->s_namelen);
				if (filldir(dirent, de->name, size, filp->f_pos, de->inode) < 0) {
					brelse(bh);
					return 0;
				}
			}
			offset += info->s_dirsize;
			filp->f_pos += info->s_dirsize;
		} while (offset < VIVA_BLOCK_SIZE && filp->f_pos < inode->i_size);
		brelse(bh);
	}
	return 0;
}
