/* -----------------------------------------------------------------
The VIVA file system - version 2.0

Implemented by Shankar Pasupathy, July 1996. (shankar@pop.uky.edu)
Modified by Vamsi Krishna, July 1997. (vamsi@cslab.uky.edu)
VIVA was developed by Eric Herrin II and Raphael Finkel.
Department of Computer Science,
University of Kentucky, Lexington 40506.
Distribution site: http://www.cslab.uky.edu/~vamsi/viva

This implementation uses code from the Minix filesystem,
(Copyright: Linus Torvalds).

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation. Read the file COPYRIGHT, included with
this distribution, for further details.
--------------------------------------------------------------------*/

/* Important functions for VIVA */
/* viva_read_super is called when a VIVA filesystem is mounted. The function
reads superblock information from disk into memory */
/* inode_getblk is another very important function - used when files are
created and written */


#ifdef MODULE
#include <linux/module.h>
#include <linux/version.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif


#include <linux/sched.h>
#include <linux/viva_fs.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/string.h>
#include <linux/stat.h>
#include <linux/locks.h>

#include <asm/system.h>
#include <asm/segment.h>
#include <asm/bitops.h>

/* Vamsi: suggested by Dr.Finkel */
int fours[256] = 
{									/* Number of zero nibbles in a byte */
	2,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1, 
	1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
	1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
	1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
	1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
	1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
	1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
	1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
	1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
	1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
	1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
	1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
	1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
	1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
	1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
	1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
};

int twos[256] =
{									/* Number of 2 consecutive zeros in a byte*/
	4,3,3,3,3,2,2,2,3,2,2,2,3,2,2,2, 
	3,2,2,2,2,1,1,1,2,1,1,1,2,1,1,1, 
	3,2,2,2,2,1,1,1,2,1,1,1,2,1,1,1, 
	3,2,2,2,2,1,1,1,2,1,1,1,2,1,1,1, 
	3,2,2,2,2,1,1,1,2,1,1,1,2,1,1,1, 
	2,1,1,1,1,0,0,0,1,0,0,0,1,0,0,0, 
	2,1,1,1,1,0,0,0,1,0,0,0,1,0,0,0, 
	2,1,1,1,1,0,0,0,1,0,0,0,1,0,0,0, 
	3,2,2,2,2,1,1,1,2,1,1,1,2,1,1,1, 
	2,1,1,1,1,0,0,0,1,0,0,0,1,0,0,0, 
	2,1,1,1,1,0,0,0,1,0,0,0,1,0,0,0, 
	2,1,1,1,1,0,0,0,1,0,0,0,1,0,0,0, 
	3,2,2,2,2,1,1,1,2,1,1,1,2,1,1,1, 
	2,1,1,1,1,0,0,0,1,0,0,0,1,0,0,0, 
	2,1,1,1,1,0,0,0,1,0,0,0,1,0,0,0, 
	2,1,1,1,1,0,0,0,1,0,0,0,1,0,0,0, 
};

int ones[256] =
{									/* Number of zeros in a byte */
	8,7,7,6,7,6,6,5,7,6,6,5,6,5,5,4, 
	7,6,6,5,6,5,5,4,6,5,5,4,5,4,4,3, 
	7,6,6,5,6,5,5,4,6,5,5,4,5,4,4,3, 
	6,5,5,4,5,4,4,3,5,4,4,3,4,3,3,2, 
	7,6,6,5,6,5,5,4,6,5,5,4,5,4,4,3, 
	6,5,5,4,5,4,4,3,5,4,4,3,4,3,3,2, 
	6,5,5,4,5,4,4,3,5,4,4,3,4,3,3,2, 
	5,4,4,3,4,3,3,2,4,3,3,2,3,2,2,1, 
	7,6,6,5,6,5,5,4,6,5,5,4,5,4,4,3, 
	6,5,5,4,5,4,4,3,5,4,4,3,4,3,3,2, 
	6,5,5,4,5,4,4,3,5,4,4,3,4,3,3,2, 
	5,4,4,3,4,3,3,2,4,3,3,2,3,2,2,1, 
	6,5,5,4,5,4,4,3,5,4,4,3,4,3,3,2, 
	5,4,4,3,4,3,3,2,4,3,3,2,3,2,2,1, 
	5,4,4,3,4,3,3,2,4,3,3,2,3,2,2,1, 
	4,3,3,2,3,2,2,1,3,2,2,1,2,1,1,0, 
};


void viva_put_inode(struct inode *inode)
{
	if (inode->i_nlink)
		return;
	inode->i_size = 0;
	viva_truncate(inode);
	viva_free_inode(inode);
}

static void viva_commit_super (struct super_block * sb,
			       struct viva_super_block * vs)
{
	mark_buffer_dirty(sb->u.viva_sb.s_sbh, 1);
	sb->s_dirt = 0;
}

void viva_write_super (struct super_block * sb)
{
	struct viva_super_block * vs;

	if (!(sb->s_flags & MS_RDONLY)) {
		vs = sb->u.viva_sb.s_vs;
		if (vs->s_state & VIVA_VALID_FS)
			vs->s_state &= ~VIVA_VALID_FS;
		viva_commit_super (sb, vs);
	}
	sb->s_dirt = 0;
}


void viva_put_super(struct super_block *sb)
{
	int count, block;
	struct buffer_head *temp_bh;

	lock_super(sb);
	if (!(sb->s_flags & MS_RDONLY)) {
		sb->u.viva_sb.s_vs->s_state = sb->u.viva_sb.s_mount_state;
/*		mark_buffer_dirty(sb->u.viva_sb.s_sbh, 1);*/
	}
	sb->s_dev = 0;
	for(count = 0 ; count < VIVA_I_MAP_SLOTS ; count++)
		brelse(sb->u.viva_sb.s_imap[count]);
	for(count = 0 ; count < VIVA_Z_MAP_SLOTS ; count++)
		brelse(sb->u.viva_sb.s_zmap[count]);
/*
	for(count = 0 ; count < VIVA_FS_MAP_SLOTS ; count++)
		brelse(sb->u.viva_sb.s_fsmap[count]);
*/
	block = sb->u.viva_sb.s_first_fsmapblock;
	for (count=0 ; count < sb->u.viva_sb.s_fsmap_blocks ; count++)
	{
		if((temp_bh = bread(sb->u.viva_sb.s_dev,block,VIVA_BLOCK_SIZE))!=NULL)
		{
			memcpy(temp_bh->b_data, sb->u.viva_sb.s_cached_fsmap 
									+ (VIVA_BLOCK_SIZE*count), VIVA_BLOCK_SIZE);
			mark_buffer_dirty(temp_bh, 1);
			block++;
		}
		else
			printk("viva_put_super: Error while unmounting the Filesystem\n"); 
	}
	for(count = 0 ; count < VIVA_FA_MAP_SLOTS ; count++)
		brelse(sb->u.viva_sb.s_famap[count]);
	brelse (sb->u.viva_sb.s_sbh);
	unlock_super(sb);
	MOD_DEC_USE_COUNT;
	return;
}

static struct super_operations viva_sops = { 
	viva_read_inode,
	NULL,
	viva_write_inode,
	viva_put_inode,
	viva_put_super,
	viva_write_super,
	viva_statfs,
	viva_remount
};

int viva_remount (struct super_block * sb, int * flags, char * data)
{
	struct viva_super_block * vs;

	vs = sb->u.viva_sb.s_vs;
	if ((*flags & MS_RDONLY) == (sb->s_flags & MS_RDONLY))
		return 0;
	if (*flags & MS_RDONLY) {
		if (vs->s_state & VIVA_VALID_FS ||
		    !(sb->u.viva_sb.s_mount_state & VIVA_VALID_FS))
			return 0;
		/* Mounting a rw partition read-only. */
		vs->s_state = sb->u.viva_sb.s_mount_state;
		mark_buffer_dirty(sb->u.viva_sb.s_sbh, 1);
		sb->s_dirt = 1;
		viva_commit_super (sb, vs);
	}
	else {
	  	/* Mount a partition which is read-only, read-write. */
		sb->u.viva_sb.s_mount_state = vs->s_state;
		vs->s_state &= ~VIVA_VALID_FS;
		mark_buffer_dirty(sb->u.viva_sb.s_sbh, 1);
		sb->s_dirt = 1;

		if (!(sb->u.viva_sb.s_mount_state & VIVA_VALID_FS))
			printk ("VIVA-fs warning: remounting unchecked fs, "
				"running fsck is recommended.\n");
		else if ((sb->u.viva_sb.s_mount_state & VIVA_ERROR_FS))
			printk ("VIVA-fs warning: remounting fs with errors, "
				"running fsck is recommended.\n");
	}
	return 0;
}


struct super_block *viva_read_super(struct super_block *s,void *data, 
				     int silent)
{
	struct buffer_head *bh;
	struct buffer_head *temp_bh;
	struct viva_super_block *vs;
	int count,dev=s->s_dev,block;
	int size, block_no, offset, segment_size, seg_num;

	MOD_INC_USE_COUNT;
	lock_super(s);
	set_blocksize(dev, VIVA_BLOCK_SIZE);
	if (!(bh = bread(dev,1,VIVA_BLOCK_SIZE))) {
		s->s_dev=0;
		unlock_super(s);
		printk("VIVA-fs: unable to read superblock\n");
		MOD_DEC_USE_COUNT;
		return NULL;
	}
	vs = (struct viva_super_block *) bh->b_data;
	s->u.viva_sb.s_dev = s->s_dev;
	s->u.viva_sb.s_vs = vs;
	s->u.viva_sb.s_sbh = bh;
	s->u.viva_sb.s_mount_state = vs->s_state;
	s->s_blocksize = VIVA_BLOCK_SIZE;
	s->s_blocksize_bits = VIVA_BLOCK_SIZE_BITS;
	s->u.viva_sb.s_ninodes = vs->s_ninodes;
	s->u.viva_sb.s_nzones = vs->s_nzones;
	s->u.viva_sb.s_imap_blocks = vs->s_imap_blocks;
	s->u.viva_sb.s_zmap_blocks = vs->s_zmap_blocks;
	s->u.viva_sb.s_fsmap_blocks = vs->s_fsmap_blocks;
	s->u.viva_sb.s_famap_blocks = vs->s_famap_blocks;
	s->u.viva_sb.s_firstdatazone = vs->s_firstdatazone;
	s->u.viva_sb.s_max_frag_blocks = vs->s_max_frag_blocks;
	s->u.viva_sb.s_frag_blocks = 0;
	s->u.viva_sb.s_last_frag_block = vs->s_firstdatazone;
	s->u.viva_sb.s_log_zone_size = vs->s_log_zone_size;
	s->s_magic = vs->s_magic;
	/* read segment information */
	s->u.viva_sb.s_inode_nsegments = vs->s_inode_nsegments;
	s->u.viva_sb.s_zone_nsegments = vs->s_zone_nsegments;
	s->u.viva_sb.s_inode_seg_size = vs->s_inode_seg_size;
	s->u.viva_sb.s_zone_seg_size = vs->s_zone_seg_size;
	s->u.viva_sb.s_inode_lastseg_size = vs->s_inode_lastseg_size;
	s->u.viva_sb.s_zone_lastseg_size = vs->s_zone_lastseg_size;

	for (count=0; count < VIVA_I_MAP_SLOTS; count++)
		s->u.viva_sb.s_imap[count] = NULL;
	for (count=0; count < VIVA_Z_MAP_SLOTS; count++)
		s->u.viva_sb.s_zmap[count] = NULL;
/*
	for (count=0; count < VIVA_FS_MAP_SLOTS; count++)
		s->u.viva_sb.s_fsmap[count] = NULL;
*/
	for (count=0; count < VIVA_FA_MAP_SLOTS; count++)
		s->u.viva_sb.s_famap[count] = NULL;
	/* read in inode and zone bitmaps */
	block=2;
	for (count=0 ; count < s->u.viva_sb.s_imap_blocks ; count++)
		if((s->u.viva_sb.s_imap[count]=bread(dev,block,VIVA_BLOCK_SIZE)) !=NULL)
			block++;
		else
			break;
	for (count=0 ; count < s->u.viva_sb.s_zmap_blocks ; count++)
		if((s->u.viva_sb.s_zmap[count]=bread(dev,block,VIVA_BLOCK_SIZE)) !=NULL)
			block++;
		else
			break;
	s->u.viva_sb.s_first_fsmapblock = block;
	for (count=0 ; count < s->u.viva_sb.s_fsmap_blocks ; count++)
		if((temp_bh = bread(dev,block,VIVA_BLOCK_SIZE))!=NULL)
		{
			memset(s->u.viva_sb.s_cached_fsmap + (VIVA_BLOCK_SIZE*count), 0,
															VIVA_FS_MAP_SIZE);
			memcpy(s->u.viva_sb.s_cached_fsmap + (VIVA_BLOCK_SIZE*count), 
											temp_bh->b_data, VIVA_BLOCK_SIZE);
			block++;
		}
		else
			break;
	for (count=0 ; count < s->u.viva_sb.s_famap_blocks ; count++)
		if((s->u.viva_sb.s_famap[count]=bread(dev,block,VIVA_BLOCK_SIZE))!=NULL)
			block++;
		else
			break;
	if (block != 2+s->u.viva_sb.s_imap_blocks+s->u.viva_sb.s_zmap_blocks
				  +s->u.viva_sb.s_fsmap_blocks+s->u.viva_sb.s_famap_blocks) 
	{
		for(count=0; count<VIVA_I_MAP_SLOTS; count++)
			brelse(s->u.viva_sb.s_imap[count]);
		for(count=0; count<VIVA_Z_MAP_SLOTS; count++)
			brelse(s->u.viva_sb.s_zmap[count]);
/*
		for(count=0; count<VIVA_FS_MAP_SLOTS; count++)
			brelse(s->u.viva_sb.s_fsmap[count]);
*/
		for(count=0; count<VIVA_FA_MAP_SLOTS; count++)
			brelse(s->u.viva_sb.s_famap[count]);
		s->s_dev=0;
		unlock_super(s);
		brelse(bh);
		printk("VIVA-fs: bad superblock or unable to read bitmaps\n");
		MOD_DEC_USE_COUNT;
		return NULL;
	}

	block_no = 0;
	offset = 0;
	segment_size = vs->s_inode_seg_size;
	for (seg_num=0; seg_num<s->u.viva_sb.s_inode_nsegments; seg_num++) 
	{
		if(seg_num == s->u.viva_sb.s_inode_nsegments-1)
			segment_size = vs->s_inode_lastseg_size;
		/*
		s->u.viva_sb.s_free_inodes[3][seg_num] = 0;
		s->u.viva_sb.s_free_inodes[2][seg_num] = 0;
		s->u.viva_sb.s_free_inodes[1][seg_num] = 0;
		s->u.viva_sb.s_free_inodes[0][seg_num] = 0;
		*/
		for(size=0; size<segment_size; size++)
		{
			if(offset/VIVA_BLOCK_SIZE)
			{
				block_no++;
				offset = 0;
			}
			if(!s->u.viva_sb.s_imap[block_no]->b_data[offset])
				s->u.viva_sb.s_free_inodes[3][seg_num]++;
			/* Vamsi: This neat idea was suggested by Dr.Finkel */
			s->u.viva_sb.s_free_inodes[2][seg_num] +=
					fours[(int)s->u.viva_sb.s_imap[block_no]->b_data[offset]];
			s->u.viva_sb.s_free_inodes[1][seg_num] +=
					twos[(int) s->u.viva_sb.s_imap[block_no]->b_data[offset]];
			s->u.viva_sb.s_free_inodes[0][seg_num] +=
					ones[(int) s->u.viva_sb.s_imap[block_no]->b_data[offset]];
			offset++;							
		}
	}

	block_no = 0;
	offset = 0;
	segment_size = vs->s_zone_seg_size;
	for (seg_num=0; seg_num<s->u.viva_sb.s_zone_nsegments; seg_num++) 
	{
		if(seg_num == s->u.viva_sb.s_zone_nsegments-1)
			segment_size = vs->s_zone_lastseg_size;
		/*
		s->u.viva_sb.s_free_zones[3][seg_num] =0;
		s->u.viva_sb.s_free_zones[2][seg_num] =0;
		s->u.viva_sb.s_free_zones[1][seg_num] =0;
		s->u.viva_sb.s_free_zones[0][seg_num] =0;
		*/
		for(size=0; size<segment_size; size++)
		{
			if(offset/VIVA_BLOCK_SIZE)
			{
				block_no++;
				offset = 0;
			}
			if(!s->u.viva_sb.s_zmap[block_no]->b_data[offset])
				s->u.viva_sb.s_free_zones[3][seg_num]++;
			s->u.viva_sb.s_free_zones[2][seg_num] += 
					fours[(int)s->u.viva_sb.s_zmap[block_no]->b_data[offset]];
			s->u.viva_sb.s_free_zones[1][seg_num] += 
					twos[(int) s->u.viva_sb.s_zmap[block_no]->b_data[offset]];
			s->u.viva_sb.s_free_zones[0][seg_num] += 
					ones[(int) s->u.viva_sb.s_zmap[block_no]->b_data[offset]];
			offset++;							
		}
	}

	if (s->s_magic == VIVA_SUPER_MAGIC) {
		s->u.viva_sb.s_dirsize = 16;
		s->u.viva_sb.s_namelen = 14;
	} else if (s->s_magic == VIVA_SUPER_MAGIC2) {
		s->u.viva_sb.s_dirsize = 32;
		s->u.viva_sb.s_namelen = 30;
	} else {
		s->s_dev = 0;
		unlock_super(s);
		brelse(bh);
		if (!silent)
			printk("VFS: Can't find a viva filesystem on dev 0x%04x.\n", dev);
		MOD_DEC_USE_COUNT;
		return NULL;
	}
	
	/* set bit 0 of the inode bitmap . This is an unfortunate
	consequence of the viva_new_inode functions returning
	0 if an error occurs */	
	set_bit(0,s->u.viva_sb.s_imap[0]->b_data);
	/* so we need to reduce segment counts for the first segment */
	for (count=0; count<=3; count++) {
		s->u.viva_sb.s_free_inodes[count][0]--; 
	}
	unlock_super(s);
	/* set up enough so that it can read an inode */
	s->s_dev = dev;
	s->s_op = &viva_sops;
	s->s_mounted = iget(s,VIVA_ROOT_INO);
	if (!s->s_mounted) {
		s->s_dev = 0;
		brelse(bh);
		printk("VIVA-fs: get root inode failed\n");
		MOD_DEC_USE_COUNT;
		return NULL;
	}
	s->u.viva_sb.s_mount_state = vs->s_state;
	if (!(s->u.viva_sb.s_mount_state & VIVA_VALID_FS))
		printk ("VIVA-fs: mounting unchecked file system, "
			"%d %d running fsck is recommended.\n",s->u.viva_sb.s_mount_state, vs->s_state);
 	else if (s->u.viva_sb.s_mount_state & VIVA_ERROR_FS)
		printk ("VIVA-fs: mounting file system with errors, "
			"running fsck is recommended.\n");
	if (!(s->s_flags & MS_RDONLY)) {
		vs->s_state &= ~VIVA_VALID_FS;
		mark_buffer_dirty(bh, 1);
		s->s_dirt = 1;
	}

	return s;
}

void viva_statfs(struct super_block *sb, struct statfs *buf, int bufsiz)
{
	struct statfs tmp;

	tmp.f_type = sb->s_magic;
	tmp.f_bsize = sb->s_blocksize;
	tmp.f_blocks = (sb->u.viva_sb.s_nzones - sb->u.viva_sb.s_firstdatazone) << sb->u.viva_sb.s_log_zone_size;
	tmp.f_bfree = viva_count_free_blocks(sb);
	tmp.f_bavail = tmp.f_bfree;
	tmp.f_files = sb->u.viva_sb.s_ninodes;
	tmp.f_ffree = viva_count_free_inodes(sb);
	tmp.f_namelen = sb->u.viva_sb.s_namelen;
	memcpy_tofs(buf, &tmp, bufsiz);
}

static struct buffer_head * inode_getblk(struct inode * inode, int nr, int create)
{
	int old_block_addr, new_block_addr;
	unsigned int goal;
	int  expand_retval;
	struct buffer_head * result;

repeat:
	old_block_addr = viva_bmap(inode->u.viva_i.i_cached_smap, nr);
	if (old_block_addr > 0) {
		/* Block exists, so just get it */
			result = getblk(inode->i_dev, old_block_addr, VIVA_BLOCK_SIZE);
	/* Is it possible that getblk blocks and some other process modifies
	block number nr ? Maybe. */
	/* If that were to happen, the following check would be necessary */
		if (old_block_addr == viva_bmap(inode->u.viva_i.i_cached_smap, nr))
			return result;
		brelse(result);
		goto repeat;
	}
	if (!create)
		return NULL;
	goal = inode->u.viva_i.i_lastblock + 1;
	if (goal < inode->i_sb->u.viva_sb.s_firstdatazone) {
		/* This will certainly happen when a file is first created */
		goal = inode->i_sb->u.viva_sb.s_firstdatazone;
	}
	new_block_addr = viva_new_block(inode->i_sb,goal,inode->u.viva_i.i_blocksInUse);
	if (new_block_addr < 0)
	{
			/* No blocks available or some error */
		printk("Vamsi: No more  blocks available or some error\n");
			return NULL;
	}
	
	result = getblk(inode->i_dev, new_block_addr, VIVA_BLOCK_SIZE);
	old_block_addr = viva_bmap(inode->u.viva_i.i_cached_smap, nr);
	if (old_block_addr > 0) 
	{
		viva_free_block(inode->i_sb,old_block_addr);
		inode->u.viva_i.i_blocksInUse--;
		brelse(result);
		goto repeat;
	}

	if(old_block_addr == -1)
	{
/*		printk(" Trouble %d ",nr);*/
		expand_retval = viva_break_hole(inode, nr, new_block_addr);
	}
	else
	{
		if ((expand_retval=viva_expand_smap(inode->u.viva_i.i_cached_smap, 
										new_block_addr, VIVA_SMAP_SIZE)) < 0) 
		{
			printk("VIVA: No space left to expand inode's sparse bitmap\n");
			viva_free_block(inode->i_sb,new_block_addr);
			return NULL;
		}
	}
	if (expand_retval > (VIVA_DIRECT_SMAP_SIZE - 1)) {
		 if (!inode->u.viva_i.i_cached_indirect_block) {
		/* need to get a free block to use as an indirect block */
		goal = inode->u.viva_i.i_lastblock;
		old_block_addr = viva_new_block(inode->i_sb,goal,0);
		if (old_block_addr < 0) {
			/* No block available */
			printk("VIVA: No block available for indirect sparse bitmap\n");
			viva_free_block(inode->i_sb,new_block_addr);
			return NULL;
		}
		inode->u.viva_i.i_cached_indirect_block = old_block_addr;
		inode->u.viva_i.indirect_smap_bh = getblk(inode->i_dev, old_block_addr, VIVA_BLOCK_SIZE);
		}
		/* the indirect block has been modified */
		inode->u.viva_i.i_iblock_dirty = 1; 
	}	
	inode->u.viva_i.i_blocksInUse++;
	inode->u.viva_i.i_lastblock = new_block_addr;
	inode->i_ctime = CURRENT_TIME;
	inode->i_dirt = 1;
	return result;
}


struct buffer_head * viva_getblk(struct inode * inode, int block, int create)
{
	if (block<0) {
		printk("viva_getblk: block<0");
		return NULL;
	}
	return inode_getblk(inode,block,create);
}

struct buffer_head * viva_bread(struct inode * inode, int block, int create)
{
	struct buffer_head * bh;

	bh = viva_getblk(inode,block,create);
	if (!bh || buffer_uptodate(bh))
		return bh;
	ll_rw_block(READ, 1, &bh);
	wait_on_buffer(bh);
	if (buffer_uptodate(bh))
		return bh;
	brelse(bh);
	return NULL;
}

void viva_read_inode(struct inode * inode)
{
	struct buffer_head * bh;
	struct viva_inode * raw_inode;
	int block, ino;
	int num_frags;

	ino = inode->i_ino;
	inode->i_op = NULL;
	inode->i_mode = 0;
	if (!ino || ino >= inode->i_sb->u.viva_sb.s_ninodes) {
		printk("Bad inode number on dev 0x%04x: %d is out of range\n",
			inode->i_dev, ino);
		return;
	}
	block = 2 + inode->i_sb->u.viva_sb.s_imap_blocks 
				+ inode->i_sb->u.viva_sb.s_zmap_blocks 
				+ inode->i_sb->u.viva_sb.s_fsmap_blocks
				+ inode->i_sb->u.viva_sb.s_famap_blocks
		    	+ (ino-1)/VIVA_INODES_PER_BLOCK;
	if (!(bh=bread(inode->i_dev,block, VIVA_BLOCK_SIZE))) {
		printk("Major problem: unable to read inode from dev 0x%04x\n",
			inode->i_dev);
		return;
	}
	raw_inode = ((struct viva_inode *) bh->b_data) +
		    (ino-1)%VIVA_INODES_PER_BLOCK;
	inode->i_mode = raw_inode->i_mode;
	inode->i_uid = raw_inode->i_uid;
	inode->i_gid = raw_inode->i_gid;
	inode->i_nlink = raw_inode->i_nlinks;
	inode->i_size = raw_inode->i_size;
	inode->i_mtime = inode->i_atime = inode->i_ctime = raw_inode->i_time;
	inode->i_blocks = inode->i_blksize = 0;

	/* Info for goal directed allocation in VIVA */
	inode->u.viva_i.i_blocksInUse = raw_inode->i_blocksInUse;
	inode->u.viva_i.i_lastblock = raw_inode->i_lastblock;
	inode->u.viva_i.i_fragsInUse = raw_inode->i_fragsInUse;
/*	printk("Vamsi: Frag_no = ");*/
	inode->u.viva_i.i_frag_block = raw_inode->i_frag_block;
	for(num_frags=0; num_frags < raw_inode->i_fragsInUse; num_frags++)
	{
		inode->u.viva_i.i_frag_no[num_frags] = raw_inode->i_frag_no[num_frags];
/*		printk("%d ",inode->u.viva_i.i_frag_no[num_frags]);*/
	}
/*	printk("\n");*/

	/* Indirect block addresses */
	inode->u.viva_i.i_cached_indirect_block = raw_inode->i_indirect_block;

	inode->u.viva_i.i_iblock_dirty = 0;

	if (S_ISCHR(inode->i_mode) || S_ISBLK(inode->i_mode))
		inode->i_rdev = (dev_t)(START_ADDRESS(raw_inode->i_smap));
	else {
		/* copy disk inode contents to memory */
		/* clear memory smap - not really necessary but safe */
		memset(inode->u.viva_i.i_cached_smap,0,VIVA_SMAP_SIZE);
		memcpy(inode->u.viva_i.i_cached_smap,raw_inode->i_smap,VIVA_DIRECT_SMAP_SIZE);
		/* If there is an indirect block, this is the time to load it in */
		/* Always use bread instead of getblk unless you know what you
		are doing with getblk. bread calls getblk anyway */
		if (raw_inode->i_indirect_block) 
		{
			if (!(inode->u.viva_i.indirect_smap_bh = bread(inode->i_dev,raw_inode->i_indirect_block, VIVA_BLOCK_SIZE))) {
					printk("VIVA: Unable to read indirect sparse bitmap1 on device 0x%04x\n", inode->i_dev);
					return;
			}
		memcpy(inode->u.viva_i.i_cached_smap+VIVA_DIRECT_SMAP_SIZE,inode->u.viva_i.indirect_smap_bh->b_data,VIVA_BLOCK_SIZE);
		}
	}
	brelse(bh);
	
	if (S_ISREG(inode->i_mode))
		inode->i_op = &viva_file_inode_operations;
	else if (S_ISDIR(inode->i_mode))
		inode->i_op = &viva_dir_inode_operations;
	else if (S_ISLNK(inode->i_mode))
		inode->i_op = &viva_symlink_inode_operations;
	else if (S_ISCHR(inode->i_mode))
		inode->i_op = &chrdev_inode_operations;
	else if (S_ISBLK(inode->i_mode))
		inode->i_op = &blkdev_inode_operations;
	else if (S_ISFIFO(inode->i_mode))
		init_fifo(inode);
}

static struct buffer_head * viva_update_inode(struct inode * inode)
{
	struct buffer_head * bh;
	struct viva_inode * raw_inode;
	struct viva_smap_entry *entry;
	int ino, block;
	struct buffer_head *ismap_bh;
	int num_frags;

	ino = inode->i_ino;
	if (!ino || ino >= inode->i_sb->u.viva_sb.s_ninodes) {
		printk("Bad inode number on dev 0x%04x: %d is out of range\n",
			inode->i_dev, ino);
		inode->i_dirt = 0;
		return 0;
	}
	ismap_bh = inode->u.viva_i.indirect_smap_bh;
	block = 2 + inode->i_sb->u.viva_sb.s_imap_blocks 
			  + inode->i_sb->u.viva_sb.s_zmap_blocks 
			  + inode->i_sb->u.viva_sb.s_fsmap_blocks
			  + inode->i_sb->u.viva_sb.s_famap_blocks
			  + (ino-1)/VIVA_INODES_PER_BLOCK;
	if (!(bh=bread(inode->i_dev, block, VIVA_BLOCK_SIZE))) {
		printk("unable to read i-node block\n");
		inode->i_dirt = 0;
		return 0;
	}
	raw_inode = ((struct viva_inode *)bh->b_data) +
		(ino-1)%VIVA_INODES_PER_BLOCK;
	raw_inode->i_mode = inode->i_mode;
	raw_inode->i_uid = inode->i_uid;
	raw_inode->i_gid = inode->i_gid;
	raw_inode->i_nlinks = inode->i_nlink;
	raw_inode->i_size = inode->i_size;
	raw_inode->i_time = inode->i_mtime;
	raw_inode->i_blocksInUse = inode->u.viva_i.i_blocksInUse;
	raw_inode->i_lastblock = inode->u.viva_i.i_lastblock;
	raw_inode->i_fragsInUse = inode->u.viva_i.i_fragsInUse;
	raw_inode->i_frag_block = inode->u.viva_i.i_frag_block;
	for(num_frags=0; num_frags < inode->u.viva_i.i_fragsInUse; num_frags++)
	{
		raw_inode->i_frag_no[num_frags] = inode->u.viva_i.i_frag_no[num_frags];
	}
	raw_inode->i_indirect_block = inode->u.viva_i.i_cached_indirect_block;
	entry = (struct viva_smap_entry *)(raw_inode->i_smap);
	if (S_ISCHR(inode->i_mode) || S_ISBLK(inode->i_mode))
		entry->start_addr = (unsigned long)(inode->i_rdev);
	else {
		/* copy sparse bitmap from memory to disk */
		memcpy(raw_inode->i_smap,inode->u.viva_i.i_cached_smap,VIVA_DIRECT_SMAP_SIZE);
		if ((inode->u.viva_i.i_cached_indirect_block)&&(inode->u.viva_i.i_iblock_dirty)) {
			/* write indirect block to disk, if it has been modified */
			memcpy(ismap_bh->b_data,inode->u.viva_i.i_cached_smap + VIVA_DIRECT_SMAP_SIZE, VIVA_BLOCK_SIZE);
		mark_buffer_dirty(ismap_bh,1);
		/* marking a buffer dirty, causes it to be written out after 
		a sync, update or bdflush */
		brelse(ismap_bh);
		}
		/* No longer dirty ! */
		inode->u.viva_i.i_iblock_dirty = 0;
	}
	inode->i_dirt=0;
	mark_buffer_dirty(bh, 1);
	return bh;
}

void viva_write_inode(struct inode * inode)
{
	struct buffer_head *bh;
	bh = viva_update_inode(inode);
	brelse(bh);
}

int viva_sync_inode(struct inode * inode)
{
	int err = 0;
	struct buffer_head *bh;

	bh = viva_update_inode(inode);
	if (bh && buffer_dirty(bh))
	{
		ll_rw_block(WRITE, 1, &bh);
		wait_on_buffer(bh);
		if (buffer_req(bh) && !buffer_uptodate(bh))
		{
			printk ("IO error syncing viva inode [%04x:%08lx]\n",
				inode->i_dev, inode->i_ino);
			err = -1;
		}
	}
	else if (!bh)
		err = -1;
	brelse (bh);
	return err;
}

static struct file_system_type viva_fs_type = {
	viva_read_super, "viva", 1, NULL
};

int init_viva_fs(void)
{
        return register_filesystem(&viva_fs_type);
}

#ifdef MODULE

int init_module(void)
{
	register_filesystem(&viva_fs_type);
	return 0;
}

void cleanup_module(void)
{
	unregister_filesystem(&viva_fs_type);
}

#endif

